<table>
    <thead>
        <tr>
            <th colspan="2">{{ instance.model_type_id }}: <i>{{ instance.instance_id }}</i></th>
        </tr>
    </thead>
    <tbody>
        {% for k, v in instance.create_info_data().items() %}
        <tr>
            <td>{{ k }}</td>
            <td>{{ v }}</td>
        </tr>
        {% endfor %}
    </tbody>
</table>
